import torch
import torch.nn as nn
from .utils import *

def rectangleMesh(x_range=(0,1), y_range=(0,1), h=0.25):
    """ 
    Input: 
    - x's range, (x_min, x_max)
    - y's range, (y_min, y_max)
    - h, mesh size, can be a tuple
    Return the element matrix (NT, 3)
    of the mesh a torch.meshgrid 
    """
    try:
        hx, hy = h[0], h[1]
    except:
        hx, hy = h, h

    # need to add h because arange is not inclusive
    xp = torch.arange(x_range[0], x_range[1]+hx, hx)
    yp = torch.arange(y_range[0], y_range[1]+hy, hy)
    nx, ny = len(xp), len(yp)

    x, y = torch.meshgrid(xp, yp)
    
    elem = []
    for j in range(ny-1):
        for i in range(nx-1):      
            a = i + j*nx
            b = (i+1) + j*nx
            d = i + (j+1)*nx
            c = (i+1) + (j+1)*nx
            elem += [[a, c, d], [b, c, a]]

    node = torch.stack([x.flatten(), y.flatten()], dim=-1)
    elem = torch.tensor(elem, dtype=torch.long)
    return node, elem

def quadpts(order=2, dtype=torch.float64):
    '''
    ported from iFEM's quadpts
    '''

    if order == 1:     # Order 1, nQuad 1
        baryCoords = [[1/3, 1/3, 1/3]]
        weight = [1]
    elif order == 2:    # Order 2, nQuad 3
        baryCoords = [[2/3, 1/6, 1/6],
                      [1/6, 2/3, 1/6],
                      [1/6, 1/6, 2/3]]
        weight = [1/3, 1/3, 1/3]
    elif order == 3:     # Order 3, nQuad 4
        baryCoords = [[1/3, 1/3, 1/3],
                      [0.6, 0.2, 0.2],
                      [0.2, 0.6, 0.2],
                      [0.2, 0.2, 0.6]]
        weight = [-27/48, 25/48, 25/48, 25/48]
    elif order == 4:     # Order 4, nQuad 6
        baryCoords = [[0.108103018168070, 0.445948490915965, 0.445948490915965],
                      [0.445948490915965, 0.108103018168070, 0.445948490915965],
                      [0.445948490915965, 0.445948490915965, 0.108103018168070],
                      [0.816847572980459, 0.091576213509771, 0.091576213509771],
                      [0.091576213509771, 0.816847572980459, 0.091576213509771],
                      [0.091576213509771, 0.091576213509771, 0.816847572980459], ]
        weight = [0.223381589678011, 0.223381589678011, 0.223381589678011,
                  0.109951743655322, 0.109951743655322, 0.109951743655322]
    return torch.tensor(baryCoords, dtype=dtype), torch.tensor(weight, dtype=dtype)

class TriMesh2D(nn.Module):
    '''
    Set up auxiliary data structures for Dirichlet boundary condition


    Combined the following routine from Long Chen's iFEM 
        - setboundary: get a boundary bool matrix according to elem
        - delmesh: delete mesh by eval()
        - auxstructure: edge-based auxiliary data structure
        - gradbasis: compute the gradient of local barycentric coords

    Input:
        - node: (N, 2)
        - elem: (NT, 3)

    Outputs:
        - edge: (NE, 2) global indexing of edges
        - elem2edge: (NT, 3) local to global indexing
        - edge2edge: (NE, 4)
          edge2elem[e,:2] are the global indexes of two elements sharing the e-th edge
          edge2elem[e,-2:] are the local indices of e to edge2elem[e,:2]
        - neighbor: (NT, 3) the local to global indices map of neighbor of elements
          neighbor[t,i] is the global index of the element opposite to the i-th vertex of the t-th element. 

    Example: the following routine gets all ifem similar data structures
        node, elem = rectangleMesh(x_range=(0,1), y_range=(0,1), h=1/16)
        T = TriMesh2D(node,elem)
        T.delete_mesh('(x>0) & (y<0)')
        T.update_auxstructure()
        T.update_gradbasis()
        node, elem = T.node, T.elem
        Dphi = T.Dlambda
        area = T.area
        elem2edgeSign = T.elem2edgeSign
        edge2elem = T.edge2elem

    Notes: 
        1. Python assigns the first appeared entry's index in unique; Matlab assigns the last appeared entry's index in unique.
        2. Matlab uses columns as natural indexing, reshape(NT, 3) in Matlab should be changed to
        reshape(3, -1).T in Python if initially the data is concatenated along axis=0 using torch.r_[].

    TODO:
        - Add Neumann boundary.
        - Change torch.bincount to torch.scatter

    '''

    def __init__(self,
                 node=None,
                 elem=None,
                 bdFlag=None,
                 dtype: torch.dtype = torch.float64,
                 ) -> None:
        super().__init__()

        self.dtype = dtype
        self.node = node.to(dtype)
        self.elem = elem
        self.bdFlag = bdFlag
        self._init_auxstruct()
        self._init_grad()

    def _init_auxstruct(self):
        elem = self.elem
        numElem = self.elem.size(0)
        numNode = self.node.size(0)

        # every edge's sign
        allEdge = torch.cat(
            [elem[:, [1, 2]], elem[:, [2, 0]], elem[:, [0, 1]]], dim=0)
        elem2edgeSign = torch.ones(3*numElem, dtype=int)
        elem2edgeSign[allEdge[:, 0] > allEdge[:, 1]] = -1
        self.elem2edgeSign = elem2edgeSign.view(3, -1).T
        allEdge, _ = torch.sort(allEdge, axis=1)
        # TODO indices in sort obj is dummy, can be used

        # edge structures
        self.edge, E2e, e2E, counts = unique(allEdge,
                                             return_counts=True,
                                             dim=0)
        self.elem2edge = e2E.view(3, -1).T

        # neighbor structures
        E2e_reverse = torch.zeros_like(E2e)
        E2e_reverse[e2E] = torch.arange(3*numElem)

        k1 = torch.div(E2e, numElem, rounding_mode='floor')
        k2 = torch.div(E2e_reverse, numElem, rounding_mode='floor')
        t1 = E2e - numElem*k1
        t2 = E2e_reverse - numElem*k2
        ix = self.isIntEdge = (counts == 2)  # interior edge indicator
        # edge to elem
        self.edge2elem = torch.stack([t1, t2, k1, k2], dim=-1)

        self.neighbor = torch.zeros((numElem, 3), dtype=int)
        ixElemLocalEdge1 = torch.stack([t1[ix], k1[ix]], dim=-1)
        ixElemLocalEdge2 = torch.stack([t2, k2], dim=-1)
        ixElemLocalEdge = torch.cat(
            [ixElemLocalEdge1, ixElemLocalEdge2], dim=0)
        ixElem = torch.cat([t2[ix], t1], dim=0)
        for i in range(3):
            ix = (ixElemLocalEdge[:, 1] == i)  # i-th edge's neighbor
            # TODO: check if bincount is necessary here
            self.neighbor[:, i] = torch.bincount(ixElemLocalEdge[ix, 0],
                                                 weights=ixElem[ix],
                                                 minlength=numElem)

        isBdEdge = (counts == 1)  # boundary edge indicator
        if self.bdFlag is None:
            self.bdFlag = isBdEdge[e2E].view(3, -1).T
        Dirichlet = self.edge[isBdEdge]
        self.isBdNode = torch.zeros(numNode, dtype=bool)
        self.isBdNode[Dirichlet.ravel()] = True
        self.freeNode = ~self.isBdNode

    def _init_grad(self):
        node, elem = self.node, self.elem

        ve1 = node[elem[:, 2]]-node[elem[:, 1]]
        ve2 = node[elem[:, 0]]-node[elem[:, 2]]
        ve3 = node[elem[:, 1]]-node[elem[:, 0]]
        area = torch.abs(0.5*(-ve3[:, 0]*ve2[:, 1] + ve3[:, 1]*ve2[:, 0]))
        gradLambda = torch.zeros((len(elem), 2, 3), dtype=self.dtype)
        # (# elem, 2-dim vector, 3 vertices)

        gradLambda[..., 2] = torch.stack(
            [-ve3[:, 1]/(2*area), ve3[:, 0]/(2*area)], dim=-1)
        gradLambda[..., 0] = torch.stack(
            [-ve1[:, 1]/(2*area), ve1[:, 0]/(2*area)], dim=-1)
        gradLambda[..., 1] = torch.stack(
            [-ve2[:, 1]/(2*area), ve2[:, 0]/(2*area)], dim=-1)
        # torch.stack with dim=-1 is equivalent to np.c_[]

        self.area = area
        self.gradLambda = gradLambda

    def get_elem2edge(self):
        return self.elem2edge

    def get_bdFlag(self):
        return self.bdFlag

    def get_edge(self):
        return self.edge

    def forward(self, x):
        return None

    def delete_mesh(self, expr=None):
        '''
        Update the mesh by deleting the eval(expr)
        '''
        assert expr is not None
        node, elem = self.node, self.elem
        center = node[elem].mean(axis=1)
        x, y = center[:, 0], center[:, 1]

        # delete element
        idx = eval(expr)
        mask = torch.ones(len(elem), dtype=bool)
        mask[idx] = False
        elem = elem[mask]

        # re-mapping the indices of vertices
        # to remove the unused ones
        isValidNode = torch.zeros(len(node), dtype=bool)
        indexMap = torch.zeros(len(node), dtype=int)

        isValidNode[elem.ravel()] = True
        self.node = node[isValidNode]

        indexMap[isValidNode] = torch.arange(len(self.node))
        self.elem = indexMap[elem]


import numpy as np
import torch
try:
    import plotly.figure_factory as ff
    import plotly.io as pio
    import plotly.graph_objects as go
except ImportError as e:
    print('Please install Plotly for showing mesh and solutions.')
import matplotlib.pyplot as plt
import matplotlib.tri as tri

def showmesh(node,elem, **kwargs):
    triangulation = tri.Triangulation(node[:,0], node[:,1], elem)
    markersize = 3000/len(node)
    if kwargs.items():
        h = plt.triplot(triangulation, 'b-h', **kwargs)
    else:
        h = plt.triplot(triangulation, 'b-h', linewidth=0.5, alpha=0.5, markersize=markersize)
    return h

def showsolution(node,elem,u,**kwargs):
    '''
    show 2D solution either of a scalar function or a vector field
    '''
    markersize = 3000/len(node)
    
    if u.ndim == 1:
        uplot = ff.create_trisurf(x=node[:,0], y=node[:,1], z=u,
                            simplices=elem,
                            colormap="Viridis", # similar to matlab's default colormap
                            showbackground=False,
                            aspectratio=dict(x=1, y=1, z=1),
                            )
        fig = go.Figure(data=uplot)

    elif u.ndim == 2 and u.shape[-1] == 2:
        assert u.shape[0] == elem.shape[0]
        u /= (np.abs(u)).max()
        center = node[elem].mean(axis=1)
        uplot = ff.create_quiver(x=center[:,0], y=center[:,1], 
                            u=u[:,0], v=u[:,1],
                            scale=.05,
                            arrow_scale=.5,
                            name='gradient of u',
                            line_width=1,
                            )

    fig = go.Figure(data=uplot)
    
    fig.update_layout(template='plotly_dark',
                    margin=dict(l=5, r=5, t=5, b=5),
                    **kwargs)
    fig.show()

def unique(x,
           sorted=False,
           return_counts=False,
           dim=None):
    """
    modified from
    https://github.com/pytorch/pytorch/issues/36748#issuecomment-619514810

    Args:
        input (Tensor): the input tensor
        sorted (bool): Whether to sort the unique elements in ascending order
            before returning as output.
        return_indices (bool): If True, also return the indices of ar (along the specified axis, if provided, or in the flattened array) that result in the unique array (added in this script).
        return_inverse (bool): Whether to also return the indices for where
            elements in the original input ended up in the returned unique list.
        return_counts (bool): Whether to also return the counts for each unique
            element.
        dim (int): the dimension to apply unique. If ``None``, the unique of the
            flattened input is returned. default: ``None``

    Returns:
        (Tensor, Tensor (optional), Tensor (optional)): A tensor or a tuple of tensors containing

            - **output** (*Tensor*): the output list of unique scalar elements.
            - **indices**: (optional) if :attr:`return_indices` is True, the indices of the first occurrences of the unique values in the original array.
            - **inverse_indices** (*Tensor*): (optional) if
              :attr:`return_inverse` is True, there will be an additional
              returned tensor (same shape as input) representing the indices
              for where elements in the original input map to in the output;
              otherwise, this function will only return a single tensor.
            - **counts** (*Tensor*): (optional) if
              :attr:`return_counts` is True, there will be an additional
              returned tensor (same shape as output or output.size(dim),
              if dim was specified) representing the number of occurrences
              for each unique value or tensor.
    """
    if return_counts:
        out, inverse, counts = torch.unique(x, 
                                    sorted=sorted, 
                                    return_inverse=True, 
                                    return_counts=True,
                                    dim=dim)
    else:
        out, inverse = torch.unique(x, 
                                    sorted=sorted, 
                                    return_inverse=False, 
                                    dim=dim)
    perm = torch.arange(inverse.size(0), dtype=inverse.dtype,
                        device=inverse.device)
    inverse, perm = inverse.flip([0]), perm.flip([0])
    # scatter_(dim, index, src)
    # Writes all values from the tensor src into self at the indices specified in the index tensor.
    indices = inverse.new_empty(out.size(0)).scatter_(0, inverse, perm)
    if return_counts:
        return out, indices, inverse, counts
    
    else:
        return out, indices, inverse


import torch
import torch.nn as nn
import torch.nn.functional as F
from .mesh2d import *

class Poisson(nn.Module):
    """
    A lightweight Poisson equation solver

    - Linear Lagrange element on triangulations

    Reference: Long Chen's iFEM library
    https://github.com/lyc102/ifem/blob/master/equation/Poisson.m
    """

    def __init__(
        self,
        domain=((0, 1), (0, 1)),
        h=1 / 4,
        quadrature_order=1,
        dtype: torch.dtype = torch.float64,
    ) -> None:
        super().__init__()
        # self.quadpts = quadpts
        self.domain = domain
        self.h = h
        self.mesh_size = int(1/h)+1
        self.quadrature = quadpts(order=quadrature_order, dtype=dtype)
        self.dtype = dtype
        self._initialize()

    def _initialize(self) -> None:
        # TODO update default TriMesh2D options (done)

        node, elem = rectangleMesh(
            x_range=self.domain[0], y_range=self.domain[1], h=self.h
        )
        self.trimesh = TriMesh2D(node, elem, dtype=self.dtype)
        # TODO set u to be parameters (done)
        self.freeNode = freeNode = self.trimesh.freeNode
        self.nDof = nDof = int(freeNode.sum())
        self.nNode = nNode = node.size(0)
        self.nElem = elem.size(0)

        self.register_parameter(
            "u", nn.Parameter(torch.zeros((nDof, 1), dtype=self.dtype))
        )
        nn.init.zeros_(self.u)
        self.register_buffer("uh", torch.zeros((nNode, 1), dtype=self.dtype))

    def _assemble(self, pde):
        node = self.trimesh.node
        elem = self.trimesh.elem
        gradPhi = self.trimesh.gradLambda
        area = self.trimesh.area
        nElem = self.nElem
        nNode = self.nNode

        phi, weight = self.quadrature

        # quadrature points
        quadPts = torch.einsum("qp, npd->qnd", phi, node[elem])

        # diffusion coefficient
        Kp = torch.stack([pde.diffusion_coeff(p) for p in quadPts], dim=0)
        K = torch.einsum("q, qn->n", weight, Kp)

        intgradPhiAgradPhi = torch.einsum(
            "n,n,ndi,ndj->nij", K, area, gradPhi, gradPhi)

        I = elem[:, :, None].expand_as(intgradPhiAgradPhi)
        J = elem[:, None, :].expand_as(intgradPhiAgradPhi)
        IJ = torch.stack([I, J])
        A = torch.sparse_coo_tensor(
            IJ.view(2, -1),
            intgradPhiAgradPhi.contiguous().view(-1),
            size=(nNode, nNode),
        )

        # right hand side
        b = torch.zeros((nNode, 1), dtype=self.dtype)

        if callable(pde.source):
            fK = torch.stack([pde.source(p) for p in quadPts], dim=0)
            bt = torch.einsum("q, qn, qp, n->np", weight, fK, phi, area)
        elif torch.is_tensor(pde.source):
            
            if pde.source.size(-1) == 1:  # (bsz, nNode, 1)
                pass
            else:  # (bsz, n, n)
                f =  F.interpolate(pde.source,
                                  size=(self.mesh_size+1, self.mesh_size+1),
                                    mode='bilinear',
                                    align_corners=True)
                fK = f.view(-1)[elem].mean(-1)
                bt = torch.einsum("q, n, qp, n->np", weight, fK, phi, area)

        b.scatter_(0, index=elem.view(-1, 1), src=bt.view(-1, 1), reduce="add")

        isBdNode = self.trimesh.isBdNode
        freeNode = self.trimesh.freeNode

        self.uh.scatter_(
            0,
            index=torch.where(isBdNode)[0].unsqueeze(-1),
            src=pde.g_D(node[isBdNode]).unsqueeze(-1),
        )
        b -= torch.sparse.mm(A, self.uh)

        self.A = A
        self.b = b

        A = A.to_dense()
        maskFreeNode = torch.outer(freeNode, freeNode)
        nDof = self.nDof
        A_int = A[maskFreeNode].view(nDof, nDof).to_sparse()
        A_int = A_int.coalesce()
        b_int = b[freeNode]
        self.b_int = b_int
        self.A_int = A_int

    def forward(self, u):
        return torch.sparse.mm(self.A_int, u)

    def solve(self, f=None) -> None:
        """
        direct solver, not working in sparse only
        """
        freeNode = self.trimesh.freeNode

        if self.A.is_sparse:
            A = self.A.to_dense()
        else:
            A = self.A.copy()
        b = self.b if f is None else f

        self.u.detach_()
        self.u = nn.Parameter(
            torch.linalg.solve(A[freeNode, :][:, freeNode], b[freeNode])
        )

    def get_u(self):
        """
        assemble u and u_g back into 1
        """
        self.uh[self.trimesh.freeNode] = self.u.detach()
        return self.uh.squeeze()

    def energy(self, u, b):
        """
        0.5*u^T A u - f*u
        """
        Au = self.forward(u)
        return 0.5 * (u.T).mm(Au) - (b.T).mm(u)



from setuptools import setup, find_packages

setup(
  name = 'torch-afem',
  packages=find_packages(include=['torch_afem', 'torch_afem.*']),
  version = '0.0.1',
  license='MIT',
  description = 'PyTorch Finite Element Method',
  long_description='PyTorch Finite Element Method',
  long_description_content_type="text/markdown",
  author = 'Shuhao Cao',
  author_email = 'scao.math@gmail.com',
  url = 'https://github.com/scaomath/torch-fem',
  keywords = ['pytorch', 'fem', 'pde'],
  install_requires=[
      'seaborn',
      'torchinfo',
      'numpy',
      'torch>=1.9.0',
      'plotly',
      'scipy',
      'psutil',
      'matplotlib',
      'tqdm',
      'PyYAML',
  ],
  classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Science/Research',
      'Topic :: Scientific/Engineering :: Mathematics',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 3.8',
  ],
)
